/*
 *
 *  C-to-C Translator
 *
 *  Rob Miller
 */

/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "ast.h"

#define Gen(n)  ((n)->analysis.gen)
#define Kill(n) ((n)->analysis.kill)
#define LiveVars(n)  ((n)->analysis.livevars)
#define DirtyVars(n) ((n)->analysis.dirtyvars)

/*************************************************************************/
/*                                                                       */
/*                        Bit vector frameworks                          */
/*                                                                       */
/*************************************************************************/

PRIVATE FlowValue EmptySet =
{FALSE,
 {(BitVector) 0}};

FlowValue TotalSet =
{FALSE,
 {~(BitVector) 0}};

#define Bit(bitnum)    (((BitVector)1)<<(bitnum))
#define SetBit(v, bitnum)    ((v).u.bitvector |= Bit(bitnum))
#define ClearBit(v, bitnum)  ((v).u.bitvector &= ~Bit(bitnum))
#define IsBitSet(v, bitnum)  (((v).u.bitvector & Bit(bitnum)) != 0)

/*
 * SetIntersection: meet operation for subset lattice
 */
inline FlowValue SetIntersection(FlowValue v1, FlowValue v2)
{
     if (v1.undefined)
	  return v2;
     else if (v2.undefined)
	  return v1;
     else {
	  FlowValue v;

	  v.undefined = FALSE;
	  v.u.bitvector = v1.u.bitvector & v2.u.bitvector;
	  return v;
     }
}

/*
 * SetUnion: meet operation for superset lattice
 */
PRIVATE inline FlowValue SetUnion(FlowValue v1, FlowValue v2)
{
     if (v1.undefined)
	  return v2;
     else if (v2.undefined)
	  return v1;
     else {
	  FlowValue v;

	  v.undefined = FALSE;
	  v.u.bitvector = v1.u.bitvector | v2.u.bitvector;
	  return v;
     }
}

/*
 * SetEqual: equality operation for bitvector frameworks
 */
PRIVATE inline Bool SetEqual(FlowValue v1, FlowValue v2)
{
     if (v1.undefined)
	  return v2.undefined;
     else
	  return !v2.undefined && v1.u.bitvector == v2.u.bitvector;
}

/*
 * GenKill: used in transfer function of a bitvector framework
 */
PRIVATE inline FlowValue GenKill(Node *node, FlowValue v)
{
     if (!v.undefined)
	  v.u.bitvector =
	      (v.u.bitvector & ~Kill(node).u.bitvector) | Gen(node).u.bitvector;
     return v;
}

/*
 * ListFind: returns bitvector with bit i set to 1 iff ith member of
 * list is node.  Useful for setting up Gen and Kill sets.
 */
PRIVATE FlowValue ListFind(List *list, Node *node)
{
     int di = 0;
     FlowValue v;

     v.undefined = FALSE;
     v.u.bitvector = 0;
     while (list) {
	  if (node == FirstItem(list))
	       SetBit(v, di);
	  ++di;
	  list = Rest(list);
     }
     return v;
}

/*
 * ListSelect: returns sublist of list, containing the ith member of
 * list iff ith bit of v is 1.  Used by a transfer function to convert
 * a bitvector flow value to persistent form during final pass.
 */
PRIVATE List *ListSelect(List *list, FlowValue v)
{
     List *new = NULL;

     while (v.u.bitvector != 0 && list != NULL) {
	  if (v.u.bitvector & 1)
	       new = ConsItem(FirstItem(list), new);
	  list = Rest(list);
	  v.u.bitvector >>= 1;
     }
     return new;
}

/* same as above, but the returned list is restricted to be a subset
 * of set.
 */
PRIVATE List *ListSelectSubset(List *list, FlowValue v, List *set)
{
     List *new = NULL;

     while (v.u.bitvector != 0 && list != NULL) {
	  if ((v.u.bitvector & 1) && FindItem(set, FirstItem(list)))
	       new = ConsItem(FirstItem(list), new);
	  list = Rest(list);
	  v.u.bitvector >>= 1;
     }
     return new;
}

/*
 *  ApplyInChunks: splits up a list into bitvector-length sublists and applies
 *                 proc to each
 */
PRIVATE void ApplyInChunks(void (*proc) (Node *, List *), Node *root, List *list)
{
     ListMarker m;
     List *chunk;

     IterateList(&m, list);

     while ((chunk = NextChunkOnList(&m, MAX_BITVECTOR_LENGTH)) != NULL)
	  proc(root, chunk);
}

/*************************************************************************/
/*                                                                       */
/*                        Live-variable analysis                         */
/*                                                                       */
/*************************************************************************/

/*
 * VariableSet: set of variables currently being analyzed 
 */
PRIVATE List *VariableSet;

/* 
 * Live variable analysis is a typical c2c analysis framework, with three
 * parts:
 * 
 *   Control: splits up variable set into manageable (32-element) 
 *             partitions for analysis
 *   Initialization: sets up tree for analysis
 *   Transfer function: transforms live-variable set as it flows through
 *                      a node, and stores results on tree in final pass
 */

/* 
 * Initialization
 */

PRIVATE void InitLVLists(Node *node)
{
     LiveVars(node) = NULL;
}

PRIVATE void InitLVGenKill(Node *node)
{
     switch (node->typ) {
	 case Id:
	      Gen(node) = ListFind(VariableSet, node->u.id.decl);
	      Kill(node) = EmptySet;
	      break;

	 case Decl:
	      Gen(node) = EmptySet;
	      Kill(node) = ListFind(VariableSet, node);
	      break;

	 case Binop:
	      if (node->u.binop.op == '=' && node->u.binop.left->typ == Id) {
		   Node *left = node->u.binop.left;

		   Gen(node) = EmptySet;
		   Kill(node) = ListFind(VariableSet, left->u.id.decl);

		   /* reset variable's gen-kill sets, since it's an lvalue and
		    * shouldn't generate a use.  (Note that this function is
		    * being executed postorder, so left has already been visited.) */
		   Gen(left) = EmptySet;
		   Kill(left) = EmptySet;
	      }
	      break;

	 case Spawn:
	      {
		   Node *receiver = node->u.spawn.receiver;
		   if (receiver && receiver->typ == Id && node->u.spawn.assign_op == '=') {
			Gen(node) = EmptySet;
			Kill(node) = ListFind(VariableSet, receiver->u.id.decl);
			Gen(receiver) = EmptySet;
			Kill(receiver) = EmptySet;
		   } else {
			Gen(node) = EmptySet;
			Kill(node) = EmptySet;
		   }

		   break;
	      }

	 case InletCall:
	      {
		   Node *receiver = node->u.inletcall.receiver;
		   if (receiver && receiver->typ == Id) {
			Gen(node) = EmptySet;
			Kill(node) = ListFind(VariableSet, receiver->u.id.decl);
			Gen(receiver) = EmptySet;
			Kill(receiver) = EmptySet;
		   } else {
			Gen(node) = EmptySet;
			Kill(node) = EmptySet;
		   }
		   /* additionally, we must add to gen all of the variables used by
		    * the inlet body.  These will be added in TransLV. -KHR
		    */
		   break;
	      }

	 default:
	      Gen(node) = EmptySet;
	      Kill(node) = EmptySet;
	      break;
     }
}

/*
 * Transfer function
 */

PRIVATE FlowValue TransLV(Node *node, FlowValue v, Point p, Bool final)
{
     switch (p) {
	 case EntryPoint:
#if 0
	      /* if you care about live variables on entry to an expression,
	       * instead of exit, then uncomment this code and comment out
	       * the duplicate code in the ExitPoint case */

	      if (final) {
		   List *livevars = ListSelect(VariableSet, v);
		   LiveVars(node) = JoinLists(livevars, LiveVars(node));
	      }
#endif
	      if (final && node->typ == Return) {
		   /* we need this to implement implicit syncs! */
		   List *livevars = ListSelect(VariableSet, v);
		   node->u.Return.livevars = JoinLists(livevars, node->u.Return.livevars);
	      }
	      return v;

	 case ExitPoint:
	      if (final) {
		   List *livevars = ListSelect(VariableSet, v);
		   LiveVars(node) = JoinLists(livevars, LiveVars(node));
	      }
	      /* do gen/kill only at ExitPoint of each expression, where Kills
	       * (assignments) presumably occur */
	      v = GenKill(node, v);

	      /* add variables used by inlet to flow value */
	      if (node->typ == InletCall) {
		   if (node->u.inletcall.name->typ == Id &&
		       node->u.inletcall.name->u.id.decl->u.decl.type->typ == Fdcl) {
			assert(node->u.inletcall.name->u.id.decl->u.decl.type->u.fdcl.proc_node != NULL);
			v = SetUnion(v, node->u.inletcall.name->u.id.decl->u.decl.type->u.fdcl.proc_node->u.proc.return_values);
			/* note: this is conservative - we aren't removing anything from
			 * the live variable set, such as variables which may be written
			 * by the inlet.  -KHR
			 */
		   } else {
			/*
			 * Inlet is referenced by pointer.  We don't know anything
			 * about the variables it uses, so add all of them to live set.
			 */
			assert(0);	/* not working yet */
			v = TotalSet;
		   }
	      }
	 default:
	      return v;
     }
}

/*
 * Control
 */

/* 
 * AnalyzeLV: performs analysis for small (<=32) set of variables 
 */
PRIVATE void AnalyzeLV(Node *root, List *vars)
{
     VariableSet = vars;

     WalkTree(root, (WalkProc) InitLVGenKill, NULL, Postorder);

     IterateDataFlow(
			 root,	/* root node */
			 EmptySet,	/* initial value: no variables live */
			 Backwards,	/* direction */
			 SetUnion,	/* meet operation */
			 SetEqual,	/* equality operation */
			 TransLV	/* transfer function */
	 );
}

/* 
 * AnalyzeLiveVariables: handles any size set of variables by splitting up
 *                        into small partitions
 */
GLOBAL void AnalyzeLiveVariables(Node *root, List *vars)
{
     /* initialize live variables list on every node to NULL */
     WalkTree(root, (WalkProc) InitLVLists, NULL, Preorder);

     /* compute flow for 32 variables at a time (to fit easily in bitvector) */
     ApplyInChunks(AnalyzeLV, root, vars);
}

/*************************************************************************/
/*                                                                       */
/*            Checking for function without return                       */
/*                                                                       */
/*************************************************************************/

#define REACHABLE        0
#define VALUE_RETURNED   1
#define NOTHING_RETURNED 2

PRIVATE FlowValue TransRF(Node *node, FlowValue v, Point p, Bool final)
{
     switch (p) {

	 case EntryPoint:
	      if (v.undefined && IsExpr(node) && node->typ != Block) {
		   if (final)
			WarningCoord(3, node->coord, "unreachable code");
		   /* make it defined so we don't get multiple warnings */
		   v = EmptySet;
	      }
	      return v;

	 case ExitPoint:
	      if (node->typ == Return) {
		   ClearBit(v, REACHABLE);
		   if (node->u.Return.expr)
			SetBit(v, VALUE_RETURNED);
		   else
			SetBit(v, NOTHING_RETURNED);
	      } else if (node->typ == Proc && final) {
		   Node *type = FunctionReturnType(node);
		   Coord end_coord;
		   assert(node->u.proc.body);
		   assert(node->u.proc.body->typ == Block);
		   end_coord = node->u.proc.body->u.Block.right_coord;

		   if (!IsVoidType(type) &&
		       (!TypeIsSint(type) || TYPE_QUALS(node->u.proc.decl->u.decl.type->u.fdcl.tq) & (T_PROCEDURE | T_INLET)) &&
		       IsBitSet(v, REACHABLE)) {
			WarningCoord(3, end_coord,
			     "control reaches end of non-void function");
			node->u.proc.needs_return = TRUE;
		   } else if (IsBitSet(v, VALUE_RETURNED) &&
			      (IsBitSet(v, NOTHING_RETURNED) || IsBitSet(v, REACHABLE)))
			WarningCoord(3, node->coord,
			"function returns both with and without a value");

		   /* mark procedure as needing an implicit return */
		   if (IsVoidType(type) && IsBitSet(v, REACHABLE)) {
			node->u.proc.needs_return = TRUE;
		   }
	      }
	      return v;
	 default:
	      return v;
     }
}

GLOBAL void AnalyzeReturnFlow(Node *proc)
{
     FlowValue v = EmptySet;
     SetBit(v, REACHABLE);
     IterateDataFlow(proc,
		     v,
		     Forwards,
		     SetUnion,
		     SetEqual,
		     TransRF);
}

/*************************************************************************/
/*                                                                       */
/*      Checking for returns that might have outstanding children        */
/*                                                                       */
/*************************************************************************/

#define CHILDREN_OUTSTANDING 1

/* 
 * Initialization
 */

PRIVATE void InitSF(Node *node, void *ignore)
{
     Gen(node) = EmptySet;
     Kill(node) = EmptySet;

     switch (node->typ) {
	 case Spawn:
	 case InletCall:
	      SetBit(Gen(node), CHILDREN_OUTSTANDING);
	      break;
	 case Sync:
	      SetBit(Kill(node), CHILDREN_OUTSTANDING);
	      break;
	 case Return:
	      /* we will add implicit sync's to these returns */
	      SetBit(Kill(node), CHILDREN_OUTSTANDING);
	 default:
	      break;
     }
}

/*
 * Transfer function
 */

PRIVATE FlowValue TransSF(Node *node, FlowValue v, Point p, Bool final)
{
     switch (p) {

	 case EntryPoint:
	      return v;

	 case ExitPoint:
	      if (final && IsBitSet(v, CHILDREN_OUTSTANDING)) {
		   if (node->typ == Return)
			node->u.Return.needs_sync = TRUE;
		   if (node->typ == Proc) {
			assert(node->u.proc.needs_return);
			node->u.proc.needs_sync = TRUE;
		   }
	      }
	      v = GenKill(node, v);
	      return v;

	 default:
	      return v;
     }
}

GLOBAL void AnalyzeSyncFlow(Node *proc)
{
     WalkTree(proc, InitSF, NULL, Preorder);
     IterateDataFlow(proc,
		     EmptySet,
		     Forwards,
		     SetUnion,
		     SetEqual,
		     TransSF);
}

/*************************************************************************/
/*                                                                       */
/*                        Dirty-variable analysis                        */
/*                                                                       */
/*************************************************************************/

/* 
 * Initialization
 */

PRIVATE void InitDVLists(Node *node)
{
     DirtyVars(node) = NULL;
}

PRIVATE void InitDVGenKill(Node *node)
{
     Gen(node) = EmptySet;
     Kill(node) = EmptySet;

     switch (node->typ) {
	 case Binop:
	      if (IsAssignmentOp(node->u.binop.op) &&
		  node->u.binop.left->typ == Id)
		   Gen(node) = ListFind(VariableSet, node->u.binop.left->u.id.decl);
	      break;

	 case Decl:
	      if (node->u.decl.init)
		   Gen(node) = ListFind(VariableSet, node);
	      break;

	 case Unary:
	      if (node->typ == Unary && node->u.unary.op == SIZEOF)
		   return;	/* don't look inside sizeof() */
	      if (node->typ == Unary && node->u.unary.op == ALIGNOF)
		   return;	/* don't look inside __alignof__() */
	      else if (IsIncDecOp(node->u.unary.op) &&
		       node->u.unary.expr->typ == Id)
		   Gen(node) = ListFind(VariableSet, node->u.binop.left->u.id.decl);
	      break;

	 case Spawn:
	      if (!node->u.spawn.in_inlet) {
		   if (node->u.spawn.receiver &&
		       node->u.spawn.receiver->typ == Id)
			Gen(node) = ListFind(VariableSet, node->u.spawn.receiver->u.id.decl);
		   /* a spawn saves all dirty, live variables */
		   Kill(node) = TotalSet;
	      }
	      break;

	 case Sync:
	      /*
	       * I am not sure what should happen here. Now I pretend that
	       * sync is a no-op for the purposes of dataflow analysis.
	       * In Cilk-n, n<4, sync saved all dirty/live variables.
	       * This is no longer true in Cilk-4.
	       * -- athena
	       * You are correct -KHR
	       */
#if 0
	      /* a sync saves all dirty, live variables */
	      Kill(node) = TotalSet;
#endif

	      break;

	 case InletCall:
	      if (node->u.inletcall.receiver &&
		  node->u.inletcall.receiver->typ == Id)
		   Gen(node) = ListFind(VariableSet, node->u.inletcall.receiver->u.id.decl);
	      /* an inletcall saves all dirty, live variables */
	      Kill(node) = TotalSet;
	      /* additionally, we must add to gen all of the variables written by
	       * the inlet body.  These will be added in TransDV. -KHR
	       */
	 default:
	      break;
     }
}

/*
 * Transfer function
 */

PRIVATE FlowValue TransDV(Node *node, FlowValue v, Point p, Bool final)
{
     switch (p) {
	 case EntryPoint:
	      if (final && node->typ == Return) {
		   /* we need this to implement implicit syncs! */
		   node->u.Return.dirtyvars =
		       JoinLists(ListSelectSubset(VariableSet,
						  v,
						node->u.Return.livevars),
				 node->u.Return.dirtyvars);
	      }
	      return v;

	 case ExitPoint:
	      if (final) {
		   /* Dirty variables are variables which need to be written to
		    * the frame during a spawn or sync.  They include any variable
		    * modified (i.e. in v) before the spawn's assignment (if any),
		    * or before the sync which is also live after the spawn or sync.
		    */
		   DirtyVars(node) = JoinLists(ListSelectSubset(VariableSet,
								v,
							 LiveVars(node)),
					       DirtyVars(node));
	      }
	      v = GenKill(node, v);

	      /* add variables written by inlet to flow value */
	      if (node->typ == InletCall) {
		   if (node->u.inletcall.name->typ == Id &&
		       node->u.inletcall.name->u.id.decl->u.decl.type->typ == Fdcl) {
			assert(node->u.inletcall.name->u.id.decl->u.decl.type->u.fdcl.proc_node != NULL);
			v = SetUnion(v, node->u.inletcall.name->u.id.decl->u.decl.type->u.fdcl.proc_node->u.proc.return_values);
		   } else {
			/*
			 * Inlet is referenced by pointer.  We don't know which
			 * variables it will dirty, so dirty all of them.
			 */
			assert(0);	/* not working yet */
			v = TotalSet;
		   }
	      }
	      return v;

	 default:
	      return v;
     }
}

/*
 * Control
 */

/* 
 * AnalyzeDV: performs analysis for small (<=32) set of variables 
 */
PRIVATE void AnalyzeDV(Node *root, List *vars)
{
     FlowValue formals = EmptySet;
     ListMarker marker;
     Node *var;

     VariableSet = vars;

     /* start with formal parameters being dirty (accurate for fast version,
      * conservative for slow version). */
     IterateList(&marker, vars);
     while (NextOnList(&marker, (GenericREF) & var)) {
	  if (NodeDeclLocation(var) == T_FORMAL_DECL)
	       formals = SetUnion(formals, ListFind(vars, var));
     }

     WalkTree(root, (WalkProc) InitDVGenKill, NULL, Postorder);

     IterateDataFlow(
			 root,	/* root node */
			 formals,	/* initial value: formal variables dirty */
			 Forwards,	/* direction */
			 SetUnion,	/* meet operation */
			 SetEqual,	/* equality operation */
			 TransDV	/* transfer function */
	 );
}

/* 
 * AnalyzeDirtyVariables: handles any size set of variables by splitting up
 *                        into small partitions
 */
GLOBAL void AnalyzeDirtyVariables(Node *root, List *vars)
{
     /* initialize dirty variables list on every node to NULL */
     WalkTree(root, (WalkProc) InitDVLists, NULL, Preorder);

     /* compute flow for 32 variables at a time (to fit easily in bitvector) */
     ApplyInChunks(AnalyzeDV, root, vars);
}

/*************************************************************************/
/*                                                                       */
/*                        Printing analysis results                      */
/*                                                                       */
/*************************************************************************/

GLOBAL void PV(List *vars)
{
     PrintVariables(stdout, vars);
     putchar('\n');
     fflush(stdout);
}

GLOBAL int PrintVariables(FILE *out, List *vars)
{
     ListMarker m;
     Node *var;
     int len = 0;

     IterateList(&m, vars);
     while (NextOnList(&m, (GenericREF) & var))
	  len += fprintf(out, "%s ", var->u.decl.name);
     return len;
}

GLOBAL int PrintAnalysis(FILE *out, Node *node)
{
     int len = 0;

     len += fprintf(out, "Live: ");
     len += PrintVariables(out, LiveVars(node));
     len += fprintf(out, "Dirty: ");
     len += PrintVariables(out, DirtyVars(node));
     return len;
}

/*************************************************************************/
/*                                                                       */
/*               Finding unaliased local variables                       */
/*                                                                       */
/*************************************************************************/

PRIVATE List *RemoveAliased(Node *node, List *vars);

/* 
 * Returns list of Decls which are potential candidates for registers.
 * Includes local, scalar variable decls appearing below node.
 * Omits arrays, structures, statics, externs, and variables whose
 * address has been taken.
 */
GLOBAL List *RegisterVariables(Node *node, List *vars)
{
     switch (node->typ) {
	 case Decl:
	      {
		   TypeQual dl = NodeDeclLocation(node);
		   TypeQual sc = NodeStorageClass(node);

		   if ((dl == T_BLOCK_DECL || dl == T_FORMAL_DECL) &&
		       (sc != T_TYPEDEF && sc != T_EXTERN && sc != T_STATIC) &&
		       (IsScalarType(NodeDataType(node)) ||
			IsPointerType(NodeDataType(node))))
			vars = ConsItem(node, vars);
		   break;
	      }

	 case Unary:
	      if (node->u.unary.op == ADDRESS)
		   /* remove any decls referenced by & */
		   vars = RemoveAliased(node->u.unary.expr, vars);
	      break;

	 default:
	      break;
     }

     /* An inlet left in the declaration for SerialVersion need not be searched,
      * because an equivalent block has already been appended at the block end.
      * But at the Analyze phase, the inlet has not yet been expanded, so it
      * should be searched.
      */
     /*  if ((node->typ != Proc) || strcmp(PhaseName, "Transform")) { KHR */

     /* don't look inside local procedures for register variables */
     if (node->typ == Proc && DECL_LOCATION(node->u.proc.decl->u.decl.tq) != T_TOP_DECL)
	  return vars;

#define CODE(n)   vars = RegisterVariables(n, vars)
     ASTWALK(node, CODE)
#undef CODE

	 return vars;
}

PRIVATE List *RemoveAliased(Node *node, List *vars)
{
     switch (node->typ) {
	 case Id:
	      assert(node->u.id.decl);
	      assert(node->u.id.decl->typ == Decl);
	      vars = RemoveItem(vars, node->u.id.decl);
	      break;

	 case Binop:
	      /* do we need this case? Records can't be register variables anyway. -KHR */
	      if (node->u.binop.op == '.')
		   vars = RemoveAliased(node->u.binop.left, vars);
	      break;

	 default:
	      break;
     }

     return vars;
}

/*************************************************************************/
/*                                                                       */
/*                        Running analysis                               */
/*                                                                       */
/*************************************************************************/

GLOBAL void AnalyzeProgram(List *program)
{
     ListMarker m;
     Node *n;

     IterateList(&m, program);
     while (NextOnList(&m, (GenericREF) & n))
	  if (n && n->typ == Proc) {
	       List *locals = RegisterVariables(n, NULL);

	       AnalyzeLiveVariables(n, locals);
	       AnalyzeDirtyVariables(n, locals);
	       AnalyzeReturnFlow(n);
	       AnalyzeSyncFlow(n);
	  }
}
